module Urbivore
  class Logger
    class << self
      include Urbivore::Interface

      attr_reader :configured_levels, :configured_level

      # worker constants
      # default to 250ms sleep (more? less?)
      DEFAULT_SLEEP_TIME        = (1.0/4.0)
      DEFAULT_TOTAL_WORKERS     = 1 
      DEFAULT_BATCH_SIZE        = 5

      # use this to enqueue entries
      def submit(entry)
        # any prep work on the object goes here...
        queue.push(entry)
      end

      def process(entry)
        processor.process(entry)
      end

      def processor(processor = nil)
        if processor
          unless processor.respond_to?(:process)
            raise Urbivore::Exceptions::ConfigurationError.new("Processor must respond_to?(:process)")
          end
          @processor = processor
        else
          @processor ||= Urbs::BasicProcessor
        end
      end
        
      def logger
        self
      end

      def workers
        @workers ||= []
      end

      def queue
        @queue ||= Queue.new
      end

      def sleep_time
        @sleep_time || DEFAULT_SLEEP_TIME
      end

      def total_workers
        @total_workers || DEFAULT_TOTAL_WORKERS
      end

      def batch_size
        @batch_size || DEFAULT_BATCH_SIZE
      end

      def sleep_time=(sleep_time)
        unless sleep_time.is_a?(Numeric) || !sleep_time
          raise Urbivore::Exceptions::ConfigurationError.new("Log.sleep_time must be a number")
        end

        # we should figure out whether access to this instance var 
        # should be locked, in this method and the getter...
        @sleep_time = sleep_time
      end

      def total_workers=(total_workers)
        unless total_workers.is_a?(Fixnum) || !total_workers
          raise Urbivore::Exceptions::ConfigurationError.new("Log.total_workers must be a number")
        end
        @total_workers = total_workers
      end

      def batch_size=(batch_size)
        unless batch_size.is_a?(Fixnum) || !batch_size
          raise Urbivore::Exceptions::ConfigurationError.new("Log.batch_size must be a number")
        end
        @batch_size = batch_size
      end

      # THREAD SAFETY...
      def restart_logging(halt = false)
        Thread.exclusive do
          reset_state
          if halt
            true
          else
            @keep_going   = true
            total_workers.times do
              workers << Thread.new do
                while @keep_going
                  sleep sleep_time
                  entries   = []
                  batch_size.times do 
                    entries << queue.pop
                  end
                  entries.each { |entry| process(entry) }
                end
              end
            end
          end
        end
      end
      
      private

      def reset_state
        @keep_going   = false
        workers.each { |thr| thr.join }
        @workers      = nil
        @logging      = false
        # close file handles, network connections and the like
        # note we leave queue alone, so it can keep getting entries
        # while we're reloading
        # processor.reset_state
      end
    end
  end
end
